# Copyright (c) 2010, 2011, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Time an action which calls other timed actions."""


from __future__ import absolute_import, print_function

__all__ = ['NestingTimedAction']

__metaclass__ = type


import datetime

from timeline.timedaction import TimedAction


class NestingTimedAction(TimedAction):
    """A variation of TimedAction which creates a nested environment.
    
    This is done by recording two 0 length timed actions in the timeline:
    one at the start of the action and one at the end, with -start and
    -stop appended to their categories.

    See `TimedAction` for more information.
    """

    def _init(self):
        self.duration = datetime.timedelta()
        self._category = self.category
        self.category = self._category + '-start'

    def finish(self):
        """Mark the TimedAction as finished."""
        end = self.timeline.start(self._category + '-stop', self.detail)
        end.duration = datetime.timedelta()



# Copyright (c) 2010, 2011, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests of the TimedAction class."""

from __future__ import absolute_import, print_function

__metaclass__ = type

import datetime

import testtools

from timeline.nestingtimedaction import NestingTimedAction
from timeline.timeline import Timeline


class TestNestingTimedAction(testtools.TestCase):

    def test_finish_adds_action(self):
        timeline = Timeline()
        action = NestingTimedAction("Sending mail", None, timeline)
        action.finish()
        self.assertEqual(1, len(timeline.actions))
        self.assertEqual(datetime.timedelta(), timeline.actions[-1].duration)

    def test__init__sets_duration(self):
        action = NestingTimedAction("Sending mail", None)
        self.assertEqual(datetime.timedelta(), action.duration)



# Copyright (c) 2010, 2011, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests of the Timeline class."""

from __future__ import absolute_import, print_function

__metaclass__ = type

import datetime

import testtools
from testtools.matchers import EndsWith

from timeline.timedaction import TimedAction
from timeline.timeline import OverlappingActionError, Timeline


class TestTimeline(testtools.TestCase):

    def test_start_returns_action(self):
        timeline = Timeline()
        action = timeline.start("Sending mail", "Noone")
        self.assertIsInstance(action, TimedAction)
        self.assertEqual("Sending mail", action.category)
        self.assertEqual("Noone", action.detail)
        self.assertEqual(None, action.duration)
        self.assertEqual(timeline, action.timeline)

    def test_can_supply_list(self):
        actions = "foo"
        timeline = Timeline(actions)
        self.assertEqual(actions, timeline.actions)

    def test_start_with_unfinished_action_fails(self):
        # A design constraint of timeline says that overlapping actions are not
        # permitted. See the Timeline docstrings.
        timeline = Timeline()
        action = timeline.start("Sending mail", "Noone")
        self.assertRaises(OverlappingActionError, timeline.start,
            "Sending mail", "Noone")

    def test_nested_start_permitted(self):
        # When explicitly requested a nested start can be done
        timeline = Timeline()
        action = timeline.start("Calling openid", "hostname", allow_nested=True)
        child = timeline.start("SQL Callback", "SELECT...")

    def test_nested_start_is_not_transitive(self):
        # nesting is explicit at each level - not inherited.
        timeline = Timeline()
        action = timeline.start("Calling openid", "hostname", allow_nested=True)
        child = timeline.start("SQL Callback", "SELECT...")
        self.assertRaises(OverlappingActionError, timeline.start,
            "Sending mail", "Noone")

    def test_multiple_nested_children_permitted(self):
        # nesting is not reset by each action that is added.
        timeline = Timeline()
        action = timeline.start("Calling openid", "hostname", allow_nested=True)
        child = timeline.start("SQL Callback", "SELECT...")
        child.finish()
        child = timeline.start("SQL Callback", "SELECT...")

    def test_multiple_starts_after_nested_group_prevented(self):
        # nesting stops being permitted when the nesting action is finished.
        timeline = Timeline()
        action = timeline.start("Calling openid", "hostname", allow_nested=True)
        action.finish()
        child = timeline.start("SQL Callback", "SELECT...")
        self.assertRaises(OverlappingActionError, timeline.start,
            "Sending mail", "Noone")

    def test_nesting_within_nesting_permitted(self):
        timeline = Timeline()
        action = timeline.start("Calling openid", "hostname", allow_nested=True)
        middle = timeline.start("Calling otherlibrary", "", allow_nested=True)
        child = timeline.start("SQL Callback", "SELECT...")

    def test_finishing_nested_within_nested_leaves_outer_nested_nesting(self):
        timeline = Timeline()
        action = timeline.start("Calling openid", "hostname", allow_nested=True)
        middle = timeline.start("Calling otherlibrary", "", allow_nested=True)
        middle.finish()
        child = timeline.start("SQL Callback", "SELECT...")

    def test_nested_actions_recorded_as_two_zero_length_actions(self):
        timeline = Timeline()
        action = timeline.start("Calling openid", "hostname", allow_nested=True)
        child = timeline.start("SQL Callback", "SELECT...")
        child.finish()
        action.finish()
        self.assertEqual(3, len(timeline.actions))
        self.assertEqual(datetime.timedelta(), timeline.actions[0].duration)
        self.assertEqual(datetime.timedelta(), timeline.actions[2].duration)

    def test_nested_category_labels(self):
        # To identify start/stop pairs '-start' and '-stop' are put onto the
        # category of nested actions:
        timeline = Timeline()
        action = timeline.start("Calling openid", "hostname", allow_nested=True)
        action.finish()
        self.assertEqual('Calling openid-start', timeline.actions[0].category)
        self.assertEqual('Calling openid-stop', timeline.actions[1].category)

    def test_start_after_finish_works(self):
        timeline = Timeline()
        action = timeline.start("Sending mail", "Noone")
        action.finish()
        action = timeline.start("Sending mail", "Noone")
        action.finish()
        self.assertEqual(2, len(timeline.actions))

    def test_baseline(self):
        timeline = Timeline()
        self.assertIsInstance(timeline.baseline, datetime.datetime)

    def test_start_sets_backtrace_by_default(self):
        timeline = Timeline()
        action = timeline.start("Sending mail", "Noone")
        self.assertNotEqual(None, action.backtrace)
        self.assertIsInstance(action.backtrace, str)
        self.assertThat(action.backtrace,
            EndsWith('    action = timeline.start("Sending mail", "Noone")\n'))

    def test_backtraces_can_be_disabled(self):
        # Passing format_stack=None to Timeline prevents backtrace gathering.
        timeline = Timeline(format_stack=None)
        action = timeline.start("Sending mail", "Noone")
        self.assertEqual(None, action.backtrace)



# Copyright (c) 2010, 2011, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests of the WSGI integration."""

from __future__ import absolute_import, print_function

import testtools

from timeline.timeline import Timeline
from timeline.wsgi import make_app


class TestMakeApp(testtools.TestCase):

    def test_app_passes_through_setting_environ_variable(self):
        environ = {'foo': 'bar'}
        expected_start_response = object()
        expected_result = object()
        def inner_app(environ, start_response):
            self.assertIsInstance(environ['timeline.timeline'], Timeline)
            self.assertEqual(expected_start_response, start_response)
            self.assertEqual('bar', environ['foo'])
            return expected_result
        app = make_app(inner_app)
        self.assertEqual(
            expected_result, app(environ, expected_start_response))



# Copyright (c) 2010, 2011, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for timeline."""

from __future__ import absolute_import, print_function

from unittest import TestLoader


def test_suite():
    test_mod_names = [
        'timeline',
        'timedaction',
        'nestingtimedaction',
        'wsgi',
        ]
    return TestLoader().loadTestsFromNames(
        ['timeline.tests.test_' + name for name in test_mod_names])



# Copyright (c) 2010, 2011, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Coordinate a sequence of non overlapping TimedActionss."""

from __future__ import absolute_import, print_function

__all__ = ['Timeline']

__metaclass__ = type

import datetime
import sys
import traceback

from pytz import utc as UTC

from timeline.nestingtimedaction import NestingTimedAction
from timeline.timedaction import TimedAction


class OverlappingActionError(Exception):
    """A new action was attempted without finishing the prior one."""
    # To make analysis easy we do not permit overlapping actions by default:
    # each action that is being timed and accrued must complete before the next
    # is started. This means, for instance, that sending mail cannot do SQL
    # queries, as both are timed and accrued. OTOH it makes analysis and
    # serialisation of timelines simpler, and for the current use cases in 
    # Launchpad this is sufficient. This constraint should not be considered
    # sacrosanct - if, in future, we desire timelines with overlapping actions,
    # as long as the OOPS analysis code is extended to generate sensible
    # reports in those situations, this can be changed. In the interim, actions
    # can be explicitly setup to permit nesting by passing allow_nested=True
    # which will cause the action to be recorded with 0 duration and a -start
    # and -stop suffix added to its category. This is potentially lossy but
    # good enough to get nested metrics and can be iterated on in the future to
    # do an actual stacked/tree model of actions - if needed.


def format_stack():
    """Format a stack like traceback.format_stack, but skip 2 frames.

    This means the stack formatting frame isn't in the backtrace itself.
    """
    return traceback.format_stack(f=sys._getframe(2))


class Timeline:
    """A sequence of TimedActions.

    This is used for collecting expensive/external actions inside Launchpad
    requests.

    :ivar actions: The actions.
    :ivar baseline: The point the timeline starts at.
    """

    def __init__(self, actions=None, format_stack=format_stack):
        """Create a Timeline.
        
        :param actions: An optional object to use to store the timeline. This
            must implement the list protocol.
        :param format_stack: The helper to use when gathering tracebacks. If
            None tracebacks are not gathered. Must return a list of strings
            for concatenation.
        """
        if actions is None:
            actions = []
        self.actions = actions
        self.baseline = datetime.datetime.now(UTC)
        self.format_stack = format_stack

    def start(self, category, detail, allow_nested=False):
        """Create a new TimedAction at the end of the timeline.

        :param category: the category for the action.
        :param detail: The detail for the action.
        :param allow_nested: If True treat this action as a nested action -
            record it twice with 0 duration, once at the start and once at the
            finish.
        :return: A TimedAction for that category and detail.
        """
        if allow_nested:
            result = NestingTimedAction(category, detail, self)
        else:
            result = TimedAction(category, detail, self)
        if self.actions and self.actions[-1].duration is None:
            raise OverlappingActionError(self.actions[-1], result)
        self.actions.append(result)
        if self.format_stack is not None:
            result.backtrace = ''.join(self.format_stack())
        return result



# Copyright (c) 2010, 2011, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests of the TimedAction class."""

from __future__ import absolute_import, print_function

__metaclass__ = type

import datetime

import testtools

from timeline.timedaction import TimedAction
from timeline.timeline import Timeline


class TestTimedAction(testtools.TestCase):

    def test_starts_now(self):
        action = TimedAction("Sending mail", None)
        self.assertIsInstance(action.start, datetime.datetime)

    def test_finish_sets_duration(self):
        action = TimedAction("Sending mail", None)
        self.assertEqual(None, action.duration)
        action.finish()
        self.assertIsInstance(action.duration, datetime.timedelta)

    def test__init__sets_category(self):
        action = TimedAction("Sending mail", None)
        self.assertEqual("Sending mail", action.category)

    def test__init__sets_detail(self):
        action = TimedAction(None, "fred.jones@example.com")
        self.assertEqual("fred.jones@example.com", action.detail)

    def test_logTuple(self):
        timeline = Timeline()
        action = TimedAction("foo", "bar", timeline)
        # Set variable for deterministic results
        action.start = timeline.baseline + datetime.timedelta(0, 0, 0, 2)
        action.duration = datetime.timedelta(0, 0, 0, 4)
        log_tuple = action.logTuple()
        self.assertEqual(5, len(log_tuple), "!= 5 elements %s" % (log_tuple,))
        # The first element is the start offset in ms.
        self.assertIsInstance(log_tuple[0], int)
        self.assertEqual(2, log_tuple[0])
        # The second element is the end offset in ms.
        self.assertIsInstance(log_tuple[1], int)
        self.assertEqual(6, log_tuple[1])
        self.assertEqual("foo", log_tuple[2])
        self.assertEqual("bar", log_tuple[3])
        # The fifth element defaults to None:
        self.assertEqual(None, log_tuple[4])

    def test_logTupleIncomplete(self):
        # Things that start and hit a timeout *may* not get recorded as
        # finishing in normal operation - they still need to generate a
        # logTuple, using now as the end point.
        timeline = Timeline()
        action = TimedAction("foo", "bar", timeline)
        # Set variable for deterministic results
        action.start = timeline.baseline + datetime.timedelta(0, 0, 0, 2)
        action._interval_to_now = lambda: datetime.timedelta(0, 0, 0, 3)
        log_tuple = action.logTuple()
        self.assertEqual(5, len(log_tuple), "!= 5 elements %s" % (log_tuple,))
        self.assertIsInstance(log_tuple[0], int)
        self.assertEqual(2, log_tuple[0])
        self.assertIsInstance(log_tuple[1], int)
        self.assertEqual(5, log_tuple[1])
        self.assertEqual("foo", log_tuple[2])
        self.assertEqual("bar", log_tuple[3])

    def test_logTupleBacktrace(self):
        # If the action has a backtrace attribute, that is placed into the
        # fifth element of logTuple.
        timeline = Timeline()
        action = TimedAction("foo", "bar", timeline)
        action.finish()
        action.backtrace = "Foo Bar"
        log_tuple = action.logTuple()
        self.assertEqual("Foo Bar", log_tuple[4])



# Copyright (c) 2010, 2011, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Time a single categorised action."""


from __future__ import absolute_import, division, print_function

__all__ = ['TimedAction']

__metaclass__ = type


import datetime

import pytz


UTC = pytz.utc


class TimedAction:
    """An individual action which has been timed.

    :ivar timeline: The timeline that this action took place within.
    :ivar start: A datetime object with tz for the start of the action.
    :ivar duration: A timedelta for the duration of the action. None for
        actions which have not completed.
    :ivar category: The category of the action. E.g. "sql".
    :ivar detail: The detail about the action. E.g. "SELECT COUNT(*) ..."
    :ivar backtrace: A backtrace for when the action was started. Useful for
        detecting code paths that cause lots of actions (usually a bad thing
        from a performance perspective).
    """

    def __init__(self, category, detail, timeline=None):
        """Create a TimedAction.

        New TimedActions have a start but no duration.

        :param category: The category for the action.
        :param detail: The detail about the action being timed.
        :param timeline: The timeline for the action.
        """
        self.start = datetime.datetime.now(UTC)
        self.duration = None
        self.category = category
        self.detail = detail
        self.timeline = timeline
        self.backtrace = None
        self._init()

    def _init(self):
        # hook for child classes.
        pass

    def __repr__(self):
        return "<%s %s[%s]>" % (self.__class__, self.category,
            self.detail[:20])

    def logTuple(self):
        """Return a 4-tuple suitable for errorlog's use."""
        offset = self._td_to_ms(self.start - self.timeline.baseline)
        if self.duration is None:
            # This action wasn't finished: pretend it has finished now
            # (even though it hasn't). This is pretty normal when action ends
            # are recorded by callbacks rather than stack-like structures. E.g.
            # storm tracers in launchpad:
            # log-trace START : starts action
            # timeout-trace START : raises 
            # log-trace FINISH is never called.
            length = self._td_to_ms(self._interval_to_now())
        else:
            length = self._td_to_ms(self.duration)
        return (offset, offset + length, self.category, self.detail,
            self.backtrace)

    def _td_to_ms(self, td):
        """Tweak on a backport from python 2.7"""
        return (td.microseconds + (
            td.seconds + td.days * 24 * 3600) * 10**6) // 10**3

    def finish(self):
        """Mark the TimedAction as finished."""
        self.duration = self._interval_to_now()

    def _interval_to_now(self):
        return datetime.datetime.now(UTC) - self.start



# Copyright (c) 2010, 2011, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""WSGI integration."""

from __future__ import absolute_import, print_function

__all__ = ['make_app']

from timeline.timeline import Timeline

def make_app(app):
    """Create a WSGI app wrapping app.

    The wrapper app injects an environ variable 'timeline.timeline' which is a
    Timeline object.
    """
    def wrapper(environ, start_response):
        environ['timeline.timeline'] = Timeline()
        return app(environ, start_response)
    return wrapper
